@@ -1,6 +1,4 @@
 {-# LANGUAGE PatternGuards #-}
-{-# LANGUAGE TypeApplications #-}
-{-# LANGUAGE MultiWayIf #-}
 
 -- This template expects CPP definitions for:
 --     MODULE_NAME = Posix | Windows
@@ -60,14 +58,14 @@
 --
 -- References:
 -- [1] <http://msdn.microsoft.com/en-us/library/windows/desktop/aa365247.aspx Naming Files, Paths and Namespaces> (Microsoft MSDN)
-#ifndef OS_PATH
+#ifndef ABSTRACT_FILEPATH
 module System.FilePath.MODULE_NAME
 #else
-module System.OsPath.MODULE_NAME.Internal
+module System.AbstractFilePath.MODULE_NAME.Internal
 #endif
     (
     -- * Separator predicates
-#ifndef OS_PATH
+#ifndef ABSTRACT_FILEPATH
     FilePath,
 #endif
     pathSeparator, pathSeparators, isPathSeparator,
@@ -76,7 +74,7 @@ module System.OsPath.MODULE_NAME.Internal
 
     -- * @$PATH@ methods
     splitSearchPath,
-#ifndef OS_PATH
+#ifndef ABSTRACT_FILEPATH
     getSearchPath,
 #endif
 
@@ -113,43 +111,36 @@ module System.OsPath.MODULE_NAME.Internal
 
 {- HLINT ignore "Use fewer imports" -}
 import Prelude (Char, Bool(..), Maybe(..), (.), (&&), (<=), not, fst, maybe, (||), (==), ($), otherwise, fmap, mempty, (>=), (/=), (++), snd)
-import Data.Bifunctor (first)
 import Data.Semigroup ((<>))
 import qualified Prelude as P
-import Data.Maybe(fromMaybe, isJust)
+import Data.Maybe(isJust)
 import qualified Data.List as L
 
-#ifndef OS_PATH
+#ifndef ABSTRACT_FILEPATH
 import Data.String (fromString)
 import System.Environment(getEnv)
-import Prelude (String, map, FilePath, Eq, IO, id, last, init, reverse, dropWhile, null, break, take, all, elem, any, span)
+import Prelude (String, map, FilePath, Eq, IO, id, last, init, reverse, dropWhile, null, break, takeWhile, take, all, elem, any, head, tail, span)
 import Data.Char(toLower, toUpper, isAsciiLower, isAsciiUpper)
-import Data.List(stripPrefix, isSuffixOf, uncons, dropWhileEnd)
+import Data.List(stripPrefix, isSuffixOf, uncons)
 #define CHAR Char
 #define STRING String
 #define FILEPATH FilePath
 #else
-import Prelude (fromIntegral, return, IO, Either(..))
-import Control.Exception ( catch, displayException, evaluate, fromException, toException, throwIO, Exception, SomeAsyncException(..), SomeException )
-import Control.DeepSeq (force)
-import System.IO.Unsafe (unsafePerformIO)
+import Prelude (fromIntegral)
+import System.AbstractFilePath.Encoding ( encodeWith )
+import GHC.IO.Encoding.Failure ( CodingFailureMode(..) )
 import qualified Data.Char as C
 #ifdef WINDOWS
-import GHC.IO.Encoding.Failure ( CodingFailureMode(..) )
 import GHC.IO.Encoding.UTF16 ( mkUTF16le )
-import qualified GHC.Foreign as GHC
 import Data.Word ( Word16 )
-import System.OsString.Data.ByteString.Short.Word16
-import System.OsString.Data.ByteString.Short ( packCStringLen )
+import System.AbstractFilePath.Data.ByteString.Short.Word16
 #define CHAR Word16
 #define STRING ShortByteString
 #define FILEPATH ShortByteString
 #else
-import GHC.IO.Encoding.Failure ( CodingFailureMode(..) )
-import qualified GHC.Foreign as GHC
 import GHC.IO.Encoding.UTF8 ( mkUTF8 )
 import Data.Word ( Word8 )
-import System.OsString.Data.ByteString.Short
+import System.AbstractFilePath.Data.ByteString.Short
 #define CHAR Word8
 #define STRING ShortByteString
 #define FILEPATH ShortByteString
@@ -269,7 +260,7 @@ splitSearchPath = f
 
 
 -- TODO for AFPP
-#ifndef OS_PATH
+#ifndef ABSTRACT_FILEPATH
 -- | Get a list of 'FILEPATH's in the $PATH variable.
 getSearchPath :: IO [FILEPATH]
 getSearchPath = fmap splitSearchPath (getEnv "PATH")
@@ -291,24 +282,13 @@ getSearchPath = fmap splitSearchPath (getEnv "PATH")
 -- > splitExtension "file.txt/boris.ext" == ("file.txt/boris",".ext")
 -- > splitExtension "file/path.txt.bob.fred" == ("file/path.txt.bob",".fred")
 -- > splitExtension "file/path.txt/" == ("file/path.txt/","")
-
--- A naive implementation would be to use @splitFileName_@ first,
--- then break filename into basename and extension, then recombine dir and basename.
--- This is way too expensive, see @splitFileName_@ comment for discussion.
---
--- Instead we speculatively split on the extension separator first, then check
--- whether results are well-formed.
 splitExtension :: FILEPATH -> (STRING, STRING)
-splitExtension x
-  -- Imagine x = "no-dots", then nameDot = ""
-  | null nameDot = (x, mempty)
-  -- Imagine x = "\\shared.with.dots\no-dots"
-  | isWindows && null (dropDrive nameDot) = (x, mempty)
-  -- Imagine x = "dir.with.dots/no-dots"
-  | any isPathSeparator ext = (x, mempty)
-  | otherwise = (init nameDot, extSeparator `cons` ext)
-  where
-    (nameDot, ext) = breakEnd isExtSeparator x
+splitExtension x = if null nameDot
+                   then (x, mempty)
+                   else (dir <> init nameDot, singleton extSeparator <> ext)
+    where
+        (dir,file)    = splitFileName_ x
+        (nameDot,ext) = breakEnd isExtSeparator file
 
 -- | Get the extension of a file, returns @\"\"@ for no extension, @.ext@ otherwise.
 --
@@ -370,7 +350,7 @@ addExtension file xs = case uncons xs of
   Just (x, _) -> joinDrive a res
     where
         res = if isExtSeparator x then b <> xs
-              else b <> (extSeparator `cons` xs)
+              else b <> singleton extSeparator <> xs
 
         (a,b) = splitDrive file
 
@@ -395,7 +375,7 @@ isExtensionOf :: STRING -> FILEPATH -> Bool
 isExtensionOf ext = \fp -> case uncons ext of
   Just (x, _)
     | x == _period -> isSuffixOf ext . takeExtensions $ fp
-  _ -> isSuffixOf (_period `cons` ext) . takeExtensions $ fp
+  _ -> isSuffixOf (singleton _period <> ext) . takeExtensions $ fp
 
 -- | Drop the given extension from a FILEPATH, and the @\".\"@ preceding it.
 --   Returns 'Nothing' if the FILEPATH does not have the given extension, or
@@ -415,7 +395,7 @@ isExtensionOf ext = \fp -> case uncons ext of
 -- > stripExtension ""     x          == Just x
 stripExtension :: STRING -> FILEPATH -> Maybe FILEPATH
 stripExtension ext path = case uncons ext of
-  Just (x, _) -> let dotExt = if isExtSeparator x then ext else _period `cons` ext
+  Just (x, _) -> let dotExt = if isExtSeparator x then ext else singleton _period <> ext
                  in stripSuffix dotExt path
   Nothing -> Just path
 
@@ -426,6 +406,7 @@ stripExtension ext path = case uncons ext of
 -- > splitExtensions "file.tar.gz" == ("file",".tar.gz")
 -- > uncurry (<>) (splitExtensions x) == x
 -- > Valid x => uncurry addExtension (splitExtensions x) == x
+-- > splitExtensions "file.tar.gz" == ("file",".tar.gz")
 splitExtensions :: FILEPATH -> (FILEPATH, STRING)
 splitExtensions x = (a <> c, d)
     where
@@ -517,21 +498,19 @@ readDriveUNC bs = case unpack bs of
 
 {- c:\ -}
 readDriveLetter :: STRING -> Maybe (FILEPATH, FILEPATH)
-readDriveLetter bs = case uncons2 bs of
-  Nothing -> Nothing
-  Just (x, c, ys)
-    | isLetter x, c == _colon -> Just $ case uncons ys of
-      Just (y, _)
-        | isPathSeparator y -> addSlash (pack [x,_colon]) ys
-      _ -> (pack [x,_colon], ys)
-    | otherwise -> Nothing
+readDriveLetter bs = case unpack bs of
+  (x:c:y:xs)
+    | c == _colon && isLetter x && isPathSeparator y -> Just $ addSlash (pack [x,_colon]) (pack (y:xs))
+  (x:c:xs)
+    | c == _colon && isLetter x -> Just (pack [x,_colon], pack xs)
+  _ -> Nothing
 
 {- \\sharename\ -}
 readDriveShare :: STRING -> Maybe (FILEPATH, FILEPATH)
 readDriveShare bs = case unpack bs of
   (s1:s2:xs) | isPathSeparator s1 && isPathSeparator s2 -> 
     let (a, b) = readDriveShareName (pack xs)
-    in Just (s1 `cons` (s2 `cons` a), b)
+    in Just (singleton s1 <> singleton s2 <> a,b)
   _ -> Nothing
 
 {- assume you have already seen \\ -}
@@ -601,15 +580,13 @@ isDrive x = not (null x) && null (dropDrive x)
 -- > splitFileName "bob" == ("./", "bob")
 -- > Posix:   splitFileName "/" == ("/","")
 -- > Windows: splitFileName "c:" == ("c:","")
--- > Windows: splitFileName "\\\\?\\A:\\fred" == ("\\\\?\\A:\\","fred")
--- > Windows: splitFileName "\\\\?\\A:" == ("\\\\?\\A:","")
 splitFileName :: FILEPATH -> (STRING, STRING)
 splitFileName x = if null path
     then (dotSlash, file)
     else (path, file)
   where
     (path, file) = splitFileName_ x
-    dotSlash = _period `cons` singleton _slash
+    dotSlash = singleton _period <> singleton _slash
 
 -- version of splitFileName where, if the FILEPATH has no directory
 -- component, the returned directory is "" rather than "./".  This
@@ -617,71 +594,11 @@ splitFileName x = if null path
 -- directory to make a valid FILEPATH, and having a "./" appear would
 -- look strange and upset simple equality properties.  See
 -- e.g. replaceFileName.
---
--- A naive implementation is
---
--- splitFileName_ fp = (drv <> dir, file)
---   where
---     (drv, pth) = splitDrive fp
---     (dir, file) = breakEnd isPathSeparator pth
---
--- but it is undesirable for two reasons:
--- * splitDrive is very slow on Windows,
--- * we unconditionally allocate 5 FilePath objects where only 2 would normally suffice.
---
--- In the implementation below we first speculatively split the input by the last path
--- separator. In the vast majority of cases this is already the answer, except
--- two exceptional cases explained below.
---
 splitFileName_ :: FILEPATH -> (STRING, STRING)
-splitFileName_ fp
-  -- If dirSlash is empty, @fp@ is either a genuine filename without any dir,
-  -- or just a Windows drive name without slash like "c:".
-  -- Run readDriveLetter to figure out.
-  | isWindows
-  , null dirSlash
-  = fromMaybe (mempty, fp) (readDriveLetter fp)
-  -- Another Windows quirk is that @fp@ could have been a shared drive "\\share"
-  -- or UNC location "\\?\UNC\foo", where path separator is a part of the drive name.
-  -- We can test this by trying dropDrive and falling back to splitDrive.
-  | isWindows
-  = case uncons2 dirSlash of
-    Just (s1, s2, bs')
-      | isPathSeparator s1
-      -- If bs' is empty, then s2 as the last character of dirSlash must be a path separator,
-      -- so we are in the middle of shared drive.
-      -- Otherwise, since s1 is a path separator, we might be in the middle of UNC path.
-      , null bs' || maybe False isIncompleteUNC (readDriveUNC dirSlash)
-      -> (fp, mempty)
-      -- This handles inputs like "//?/A:" and "//?/A:foo"
-      | isPathSeparator s1
-      , isPathSeparator s2
-      , Just (s3, s4, bs'') <- uncons2 bs'
-      , s3 == _question
-      , isPathSeparator s4
-      , null bs''
-      , Just (drive, rest) <- readDriveLetter file
-      -> (dirSlash <> drive, rest)
-    _ -> (dirSlash, file)
-  | otherwise
-    = (dirSlash, file)
+splitFileName_ fp = (drv <> dir, file)
   where
-    (dirSlash, file) = breakEnd isPathSeparator fp
-    dropExcessTrailingPathSeparators x
-      | hasTrailingPathSeparator x
-      , let x' = dropWhileEnd isPathSeparator x
-      , otherwise = if | null x' -> singleton (last x)
-                       | otherwise -> addTrailingPathSeparator x'
-      | otherwise = x
-
-    -- an "incomplete" UNC is one without a path (but potentially a drive)
-    isIncompleteUNC (pref, suff) = null suff && not (hasPenultimateColon pref)
-
-    -- e.g. @//?/a:/@ or @//?/a://@, but not @//?/a:@
-    hasPenultimateColon pref
-      | hasTrailingPathSeparator pref
-      = maybe False (maybe False ((== _colon) . snd) . unsnoc . fst) . unsnoc . dropExcessTrailingPathSeparators $ pref
-      | otherwise = False
+    (drv, pth) = splitDrive fp
+    (dir, file) = breakEnd isPathSeparator pth
 
 -- | Set the filename.
 --
@@ -695,7 +612,6 @@ replaceFileName x y = a </> y where (a,_) = splitFileName_ x
 --
 -- > dropFileName "/directory/file.ext" == "/directory/"
 -- > dropFileName x == fst (splitFileName x)
--- > isPrefixOf (takeDrive x) (dropFileName x)
 dropFileName :: FILEPATH -> FILEPATH
 dropFileName = fst . splitFileName
 
@@ -748,7 +664,9 @@ hasTrailingPathSeparator x
 
 
 hasLeadingPathSeparator :: FILEPATH -> Bool
-hasLeadingPathSeparator = maybe False (isPathSeparator . fst) . uncons
+hasLeadingPathSeparator x
+  | null x = False
+  | otherwise = isPathSeparator $ head x
 
 
 -- | Add a trailing file path separator if one is not already present.
@@ -812,7 +730,7 @@ combineAlways a b | null a = b
                       [a1, a2] | isWindows
                                , isLetter a1
                                , a2 == _colon -> a <> b
-                      _ -> a <> (pathSeparator `cons` b)
+                      _ -> a <> singleton pathSeparator <> b
 
 
 -- | Combine two paths with a path separator.
@@ -975,21 +893,11 @@ makeRelative root path
       where (a, b) = break isPathSeparator $ dropWhile isPathSeparator x
 
     -- on windows, need to drop '/' which is kind of absolute, but not a drive
-    dropAbs x
-      | Just (hd, tl) <- uncons x
-      , isPathSeparator hd
-      , not (hasDrive x)
-      = tl
-      | otherwise
-      = dropDrive x
+    dropAbs x | not (null x) && isPathSeparator (head x) && not (hasDrive x) = tail x
+    dropAbs x = dropDrive x
 
-    takeAbs x
-      | Just (hd, _) <- uncons x
-      , isPathSeparator hd
-      , not (hasDrive x)
-      = singleton pathSeparator
-      | otherwise
-      = map (\y -> if isPathSeparator y then pathSeparator else toLower y) $ takeDrive x
+    takeAbs x | not (null x) && isPathSeparator (head x) && not (hasDrive x) = singleton pathSeparator
+    takeAbs x = map (\y -> if isPathSeparator y then pathSeparator else toLower y) $ takeDrive x
 
 -- | Normalise a file
 --
@@ -1009,7 +917,6 @@ makeRelative root path
 -- > Posix:   normalise "./bob/fred/" == "bob/fred/"
 -- > Windows: normalise "c:\\file/bob\\" == "C:\\file\\bob\\"
 -- > Windows: normalise "c:\\" == "C:\\"
--- > Windows: normalise "c:\\\\\\\\" == "C:\\"
 -- > Windows: normalise "C:.\\" == "C:"
 -- > Windows: normalise "\\\\server\\test" == "\\\\server\\test"
 -- > Windows: normalise "//server/test" == "\\\\server\\test"
@@ -1040,7 +947,7 @@ normalise filepath =
            then singleton _period
            else joinDrive d p
 
-    addPathSeparator = isDirPath pth
+    addPathSeparator = isDirPath filepath
       && not (hasTrailingPathSeparator result)
       && not (isRelativeDrive drv)
 
@@ -1061,12 +968,9 @@ normaliseDrive :: FILEPATH -> FILEPATH
 normaliseDrive bs
   | null bs = mempty
   | isPosix = pack [pathSeparator]
-  | Just (drv, _) <- readDriveLetter x2
-                  = case unpack drv of
-                      (x:_:[]) -> pack [toUpper x, _colon]
-                      (x:_) -> pack [toUpper x, _colon, pathSeparator]
-                      _ -> P.error "impossible"
-  | otherwise = x2
+  | otherwise = if isJust $ readDriveLetter x2
+         then map toUpper x2
+         else x2
     where
         x2 = map repSlash bs
         repSlash x = if isPathSeparator x then pathSeparator else x
@@ -1148,7 +1052,7 @@ makeValid path
   | isPosix = map (\x -> if x == _nul then _underscore else x) path
   | isJust (readDriveShare drv) && all isPathSeparator drv = take 2 drv <> fromString "drive"
   | isJust (readDriveUNC drv) && not (hasTrailingPathSeparator drv) =
-      makeValid (drv <> (pathSeparator `cons` pth))
+      makeValid (drv <> singleton pathSeparator <> pth)
   | otherwise = joinDrive drv $ validElements $ validCHARs pth
 
   where
@@ -1206,12 +1110,21 @@ isRelativeDrive x =
 isAbsolute :: FILEPATH -> Bool
 isAbsolute = not . isRelative
 
-#ifndef OS_PATH
+#ifndef ABSTRACT_FILEPATH
 
 -----------------------------------------------------------------------------
+-- dropWhileEnd (>2) [1,2,3,4,1,2,3,4] == [1,2,3,4,1,2])
+-- Note that Data.List.dropWhileEnd is only available in base >= 4.5.
+dropWhileEnd :: (a -> Bool) -> [a] -> [a]
+dropWhileEnd p = reverse . dropWhile p . reverse
+
+-- takeWhileEnd (>2) [1,2,3,4,1,2,3,4] == [3,4])
+takeWhileEnd :: (a -> Bool) -> [a] -> [a]
+takeWhileEnd p = reverse . takeWhile p . reverse
+
 -- spanEnd (>2) [1,2,3,4,1,2,3,4] = ([1,2,3,4,1,2], [3,4])
 spanEnd :: (a -> Bool) -> [a] -> ([a], [a])
-spanEnd p = L.foldr (\x (pref, suff) -> if null pref && p x then (pref, x : suff) else (x : pref, suff)) ([], [])
+spanEnd p xs = (dropWhileEnd p xs, takeWhileEnd p xs)
 
 -- breakEnd (< 2) [1,2,3,4,1,2,3,4] == ([1,2,3,4,1],[2,3,4])
 breakEnd :: (a -> Bool) -> [a] -> ([a], [a])
@@ -1223,16 +1136,11 @@ breakEnd p = spanEnd (not . p)
 stripSuffix :: Eq a => [a] -> [a] -> Maybe [a]
 stripSuffix xs ys = reverse P.<$> stripPrefix (reverse xs) (reverse ys)
 
-cons :: a -> [a] -> [a]
-cons = (:)
 
 unsnoc :: [a] -> Maybe ([a], a)
-unsnoc = L.foldr (\x -> Just . maybe ([], x) (first (x :))) Nothing
+unsnoc [] = Nothing
+unsnoc xs = Just (init xs, last xs)
 
-uncons2 :: [a] -> Maybe (a, a, [a])
-uncons2 [] = Nothing
-uncons2 [_] = Nothing
-uncons2 (x : y : zs) = Just (x, y, zs)
 
 _period, _quotedbl, _backslash, _slash, _question, _U, _N, _C, _colon, _semicolon, _US, _less, _greater, _bar, _asterisk, _nul, _space, _underscore :: Char
 _period = '.'
@@ -1270,32 +1178,12 @@ snoc :: String -> Char -> String
 snoc str = \c -> str <> [c]
 
 #else
--- | Like 'try', but rethrows async exceptions.
-trySafe :: Exception e => IO a -> IO (Either e a)
-trySafe ioA = catch action eHandler
- where
-  action = do
-    v <- ioA
-    return (Right v)
-  eHandler e
-    | isAsyncException e = throwIO e
-    | otherwise = return (Left e)
-
-isAsyncException :: Exception e => e -> Bool
-isAsyncException e =
-    case fromException (toException e) of
-        Just (SomeAsyncException _) -> True
-        Nothing -> False
 #ifdef WINDOWS
 fromString :: P.String -> STRING
-fromString str = P.either (P.error . P.show) P.id $ unsafePerformIO $ do
-  r <- trySafe @SomeException $ GHC.withCStringLen (mkUTF16le ErrorOnCodingFailure) str $ \cstr -> packCStringLen cstr
-  evaluate $ force $ first displayException r
+fromString = P.either (P.error . P.show) P.id . encodeWith (mkUTF16le ErrorOnCodingFailure)
 #else
 fromString :: P.String -> STRING
-fromString str = P.either (P.error . P.show) P.id $ unsafePerformIO $ do
-  r <- trySafe @SomeException $ GHC.withCStringLen (mkUTF8 ErrorOnCodingFailure) str $ \cstr -> packCStringLen cstr
-  evaluate $ force $ first displayException r
+fromString = P.either (P.error . P.show) P.id . encodeWith (mkUTF8 ErrorOnCodingFailure)
 #endif
 
 _a, _z, _A, _Z, _period, _quotedbl, _backslash, _slash, _question, _U, _N, _C, _colon, _semicolon, _US, _less, _greater, _bar, _asterisk, _nul, _space, _underscore :: CHAR
