@@ -9,28 +9,28 @@
 --
 -- An implementation of platform specific short 'OsString', which is:
 --
--- 1. on windows wide char bytes (@[Word16]@)
--- 2. on unix char bytes (@[Word8]@)
+-- 1. on windows UTF16 data
+-- 2. on unix UTF8 data
 --
--- It captures the notion of syscall specific encoding (or the lack thereof) to avoid roundtrip issues
--- and memory fragmentation by using unpinned byte arrays. Bytes are not touched or interpreted.
+-- It captures the notion of syscall specific encoding to avoid roundtrip issues
+-- and memory fragmentation by using unpinned byte arrays.
 module System.OsString
   (
   -- * String types
     OsString
 
-  -- * OsString construction
-  , encodeUtf
-  , encodeWith
-  , encodeFS
+  -- * String construction
+  , toOsStringUtf
+  , toOsStringEnc
+  , toOsStringFS
   , osstr
-  , pack
+  , packOsString
 
-  -- * OsString deconstruction
-  , decodeUtf
-  , decodeWith
-  , decodeFS
-  , unpack
+  -- * String deconstruction
+  , fromOsStringUtf
+  , fromOsStringEnc
+  , fromOsStringFS
+  , unpackOsString
 
   -- * Word types
   , OsChar
@@ -46,15 +46,15 @@ where
 import System.OsString.Internal
     ( unsafeFromChar
     , toChar
-    , encodeUtf
-    , encodeWith
-    , encodeFS
+    , fromOsStringUtf
+    , fromOsStringEnc
+    , fromOsStringFS
     , osstr
-    , pack
-    , decodeUtf
-    , decodeWith
-    , decodeFS
-    , unpack
+    , packOsString
+    , toOsStringUtf
+    , toOsStringEnc
+    , toOsStringFS
+    , unpackOsString
     )
 import System.OsString.Internal.Types
     ( OsString, OsChar )
